#include <opencv/cv.h>
#include <opencv2/imgproc/imgproc.hpp>
#include "FlowField.h"
#include "tools.h"
#include <iostream>

void writeHelp()
{
    std::cout << "  Usage: ./FlowFields imageFilename1 imageFilename2 [options] \n\n"
              << "  Options are:\n"
              << "  -r <int>(6)		Patchradius. Patchsize for matching is p*2+1 \n"
              << "  -r2 <int>(5)		Patchradius2 (for 2. consistency check). Set to 0 for no 2. consistency check. \n"
              << "  -k <int>(3)		k parameter from paper. There is probably a logarithmic correlation between image size and optimal k i.e use larger k for larger images. \n"
              << "  -e <int>(4)		e parameter from paper. Valid range: 1-9.\n"
              << "  -s <int>(50)		s parameter from paper.\n"
              << "  -epsilon <float>(5)	epsilon parameter from paper. Even if you do not want to tune parameters consider to test epsilon = 1 as alternative setting.\n"
#ifdef ENABLE_SIFTFLOW
              << "  -SIFTFlow		Use SIFTFlow data term with S = "<< S_FROM_PAPER<<" instead of census transform. Note: to change S you have to recompile this application. S_2 != S is currently not supported. If you require it for an experiment please contact us."<< "\n"
#endif
              << "  -ouff <File>		Output raw/unfiltered Flow Field to specified file. File is saved in the Middlebury format. \n"
              << "  -offf <File>		Output outlier filtered Flow Field to specified file. Filtered flows are set to the value infinity. File is saved in the Middlebury format.\n"
              << "  -osm <File>		Output sparse matches to specified file (Section 3.5 in paper). These can be processed by Epicflow (http://lear.inrialpes.fr/src/epicflow/) to create optical flow.\n\n"
              << "  If neither -ouff nor -offf nor -osm is specified all 3 outputs will be written using meaningful filenames (see source).\n"
              << "  Please read our paper and paper supplementary material to get an idea on how to set the parameters. The SIFTFlow data term e.g. requires usually smaller r and r_2 than the census transform data term.\n"
	      <<  std::endl;
}

// Example code that shows how to use the FlowFields. Adapt TODO fields before using:
int main ( int argc,char *argv[] )
{
    if ( argc < 3 )
    {
        std::cout << "Not enough arguments."<<std::endl;
        writeHelp();
        return 1;
    }
    std::string filename1 = argv[1];
    std::string filename2 = argv[2];

    // Read the paper and the paper supplementary material to get an idea on how to set the parameters.
    // The SIFTFlow data term e.g. requires usually smaller r and r_2 than the census transform data term.
    int patchradius = 6, patchradius2 = 5, k_from_paper = 3, e_from_paper = 4, s_from_paper = 50;
    float epsilon_from_paper = 5;
    bool use_SIFTFlow = false;
    // #define S_FROM_PAPER 3;   // S has to be defined at compile time, it is defined in the cmake file

    std::string matchesFilename,unfilteredFilename, filteredFilename;
    int cArg = 3;

#define tryGetInt(x,y) if(!strcmp(argv[cArg],x)){ y = atoi(argv[(++cArg)++]);  continue;}
#define tryGetFlt(x,y) if(!strcmp(argv[cArg],x)){ y = atof(argv[(++cArg)++]);  continue;}
#define tryGetStr(x,y) if(!strcmp(argv[cArg],x)){ y =      argv[(++cArg)++] ;  continue;}

    while ( cArg < argc )
    {
        tryGetInt ( "-r",patchradius );
        tryGetInt ( "-r2",patchradius2 );
        tryGetInt ( "-k",k_from_paper );
        tryGetInt ( "-e",e_from_paper );
        tryGetInt ( "-s",s_from_paper );
        tryGetFlt ( "-epsilon",epsilon_from_paper );
        tryGetStr ( "-osm",matchesFilename );
        tryGetStr ( "-ouff",unfilteredFilename );
        tryGetStr ( "-offf",filteredFilename );
        if ( !strcmp ( argv[cArg],"-SIFTFlow" ) )
        {
            use_SIFTFlow = true;
            cArg++;
            continue;
        }
        std::cout << "Unknown argument: "<<  argv[cArg]  << std::endl;
        writeHelp();
        exit ( 0 );
    }

    if ( matchesFilename.empty() && unfilteredFilename.empty() && filteredFilename.empty() )
    {
        std::cout << "Warning: No output file was set thus the unfiltered and filtered Flow Fields and the Epic matches are saved under predefined filenames." << std::endl;
        matchesFilename = filename1+".match";
        unfilteredFilename = filename1+"_unfiltered.flo";
        filteredFilename = filename1+"_filtered.flo";
    }

    cv::Mat3f image1 =  cv::imread ( filename1,1 );
    // For census we want to have the images in lab space (see paper):
    image1/=255.f;
    cv::cvtColor ( image1,image1,CV_RGB2Lab );

    cv::Mat3f image2 =  cv::imread ( filename2,1 );
    image2/=255.f;
    cv::cvtColor ( image2,image2,CV_RGB2Lab );

    FlowField * ff;

    if ( ! use_SIFTFlow )
    {
        //---Flow Field with Census transform:---//
        ff  = new FlowField ( image1,image2, patchradius, patchradius2,Census,cv::Mat(),cv::Mat(),k_from_paper );
    }
    else
    {
#ifdef ENABLE_SIFTFLOW
        //---Flow Field with SIFTFlow. Note that a S_2 != S (like used in the paper) is currently not supported within a single Flow Field object. If you require it for an experiment please contact us---//

        // For SIFTFlow we need black and white images:
        cv::Mat1b image1BW = cv::imread ( filename1,0 );
        cv::Mat1b image2BW = cv::imread ( filename2,0 );

        _cvMatf ( S_FROM_PAPER ) sf1,sf2;
        _cvMatf ( 128 ) s1 =  SIFT_ ( image1BW );
        _cvMatf ( 128 ) s2 =  SIFT_ ( image2BW );
        PCA ( s1,s2,sf1,sf2 );

	ff = new FlowField ( image1,image2, patchradius, patchradius2,SED,sf1,sf2,k_from_paper );
#else
	std::cout << "Error: compiled without SIFTFlow support!"<<std::endl;
	exit(1);
#endif
    }

    // Returns the unfiltered Flow Field.
    cv::Mat2f FlowFieldUnfiltered = ff->getFlowField ( 1 );
    if ( !unfilteredFilename.empty() ) saveFlowField ( FlowFieldUnfiltered, unfilteredFilename );

    if ( !filteredFilename.empty() ||  !matchesFilename.empty() )
    {
        //Generates Filtered Flow Field and returns it. It can after generation also be accessed with ff.getFlowField ( 0 ).
        cv::Mat2f FlowFieldFiltered =  ff->genFiltered ( epsilon_from_paper, s_from_paper, FlowField::CONSI_STD );
        if ( !filteredFilename.empty() ) saveFlowField ( FlowFieldFiltered, filteredFilename );

        //To use EpicFlow for the final flow you can save the matches in a file and use the EpicFlow executable (http://lear.inrialpes.fr/src/epicflow/) to create the final flow:
        if ( !matchesFilename.empty() )  saveMatches ( ff->getEpicMatches ( e_from_paper ), matchesFilename );
    }
    delete ff;
    return 0;
}
